﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.NCA {
  public class NeighborhoodComponentsAnalysis {

    public static INCAModel Train(IClassificationProblemData data, int k, int reduceDimensions) {
      var instances = data.TrainingIndices.Count();
      var attributes = data.AllowedInputVariables.Count();
      var matrix = new double[attributes * reduceDimensions];
      // TODO: make some more clever initialization of matrix
      for (var i = 0; i < attributes; i++)
        for (var j = 0; j < reduceDimensions; j++)
          matrix[j * attributes + i] = 1;

      alglib.mincgstate state;
      alglib.mincgreport rep;

      // first run
      alglib.mincgcreate(matrix, out state);
      alglib.mincgsetcond(state, 0.0000000001, 0, 0, 0);
      alglib.mincgoptimize(state, Gradient, null, new OptimizationInfo(data, reduceDimensions));
      alglib.mincgresults(state, out matrix, out rep);

      var transformationMatrix = new double[attributes, reduceDimensions];
      for (var i = 0; i < attributes; i++)
        for (var j = 0; j < reduceDimensions; j++)
          transformationMatrix[i, j] = matrix[j * attributes + i];

      var transformedTrainingset = new double[instances, reduceDimensions];
      var rowCount = 0;
      for (var r = data.TrainingPartition.Start; r < data.TrainingPartition.End; r++) {
        if (r >= data.TestPartition.Start && r < data.TestPartition.End) continue;
        var j = 0;
        foreach (var v in data.AllowedInputVariables) {
          var val = data.Dataset.GetDoubleValue(v, r);
          for (var i = 0; i < transformedTrainingset.GetLength(1); i++)
            transformedTrainingset[rowCount, i] += val * transformationMatrix[j, i];
          j++;
        }
        rowCount++;
      }

      return new NCAModel(transformedTrainingset, transformationMatrix, k, data.TargetVariable, data.AllowedInputVariables,
        data.Dataset.GetDoubleValues(data.TargetVariable)
          .Select((v, i) => new { I = i, V = v })
          .Where(x => x.I >= data.TrainingPartition.Start && x.I < data.TrainingPartition.End
              && !(x.I >= data.TestPartition.Start && x.I < data.TestPartition.End))
          .Select(x => x.V).ToArray());
    }

    private static void Gradient(double[] A, ref double func, double[] grad, object obj) {
      var info = (OptimizationInfo)obj;
      var instances = info.ProblemData.TrainingIndices.ToArray();
      var distances = new double[instances.Length, instances.Length];
      for (int i = 0; i < instances.Length - 1; i++) {
        for (int j = i + 1; j < instances.Length; j++) {
          distances[i, j] = VectorLength(VectorSubtract(
            Ax(A, GetRow(info.ProblemData, instances[i])),
            Ax(A, GetRow(info.ProblemData, instances[j]))));
          distances[j, i] = distances[i, j];
        }
      }
      var probabilities = new double[instances.Length, instances.Length];
      for (int i = 0; i < instances.Length - 1; i++) {
        var denom = 0.0;
        for (int k = 0; k < instances.Length; k++) {
          if (k == i) continue;
          denom += Math.Exp(-(distances[i, k] * distances[i, k]));
        }
        for (int j = i + 1; j < instances.Length; j++) {
          probabilities[i, j] = Math.Exp(-(distances[i, j] * distances[i, j])) / denom;
        }
      }

      var target = info.ProblemData.TargetVariable;
      var ds = info.ProblemData.Dataset;
      var pi = new double[instances.Length];
      func = 0.0;
      for (int i = 0; i < instances.Length; i++) {
        var classI = ds.GetDoubleValue(target, instances[i]);
        pi[i] = 0.0;
        for (int j = 0; j < instances.Length; j++) {
          if (i == j) continue;
          var classJ = ds.GetDoubleValue(target, instances[j]);
          if (classI.IsAlmost(classJ)) {
            func += probabilities[i, j];
            pi[i] += probabilities[i, j];
          }
        }
      }

      var sum = GetSum(probabilities, pi, info.ProblemData, instances);
      for (int i = 0; i < A.Length; i++) {
        grad[i] = 2.0 * A[i] * sum;
      }
    }

    private static double[,] GetSum(double[,] p, double[] pi, IClassificationProblemData data, int[] instances) {
      var target = data.TargetVariable;
      var sum = new double[pi.Length, pi.Length];
      for (int i = 0; i < pi.Length; i++) {
        var classI = data.Dataset.GetDoubleValue(target, instances[i]);
        var sumK = 0.0;
        for (int k = 0; k < pi.Length; k++) {
          if (i == k) continue;
          sumK += p[i, k] * MatrixMultiply(VectorSubtract(GetRow(data, i), GetRow(data, k)), VectorSubtract(GetRow(data, i), GetRow(data, k)));
        }
        var sumJ = 0.0;
        for (int j = 0; j < pi.Length; j++) {
          if (i == j) continue;
          var classJ = data.Dataset.GetDoubleValue(target, instances[j]);
          if (classI.IsAlmost(classJ)) sumJ += p[i, j] * ScalarProduct(VectorSubtract(GetRow(data, i), GetRow(data, j)), VectorSubtract(GetRow(data, i), GetRow(data, j)));
        }
        sum += pi[i] * sumK - sumJ;
      }
      return sum;
    }

    private static double[,] MatrixMultiply(IEnumerable<double> vec1, IEnumerable<double> vec2) {
      var enumVec1 = vec1.GetEnumerator();
      var enumVec2 = vec2.GetEnumerator();
      bool f1, f2;
      while (true) {
        f1 = enumVec1.MoveNext();
        f2 = enumVec2.MoveNext();
        if (!f1 && f2 || f1 && !f2) throw new ArgumentException("vectors are of unequal length.");
        if (!f1 && !f2) yield break;
        yield return enumVec1.Current - enumVec2.Current;
      }
    }

    private static IEnumerable<double> Ax(IEnumerable<double> A, IEnumerable<double> x) {
      var enumA = A.GetEnumerator();
      var enumX = x.GetEnumerator();
      var sum = 0.0;
      while (true) {
        if (!enumA.MoveNext()) {
          yield return sum;
          yield break;
        } else if (!enumX.MoveNext()) {
          yield return sum;
          enumX = x.GetEnumerator();
          sum = 0.0;
        } else {
          sum += enumA.Current * enumX.Current;
        }
      }
    }

    private static IEnumerable<double> GetRow(IClassificationProblemData data, int row) {
      return data.AllowedInputVariables.Select(v => data.Dataset.GetDoubleValue(v, row));
    }

    private static IEnumerable<double> VectorSubtract(IEnumerable<double> vec1, IEnumerable<double> vec2) {
      var enumVec1 = vec1.GetEnumerator();
      var enumVec2 = vec2.GetEnumerator();
      bool f1, f2;
      while (true) {
        f1 = enumVec1.MoveNext();
        f2 = enumVec2.MoveNext();
        if (!f1 && f2 || f1 && !f2) throw new ArgumentException("vectors are of unequal length.");
        if (!f1 && !f2) yield break;
        yield return enumVec1.Current - enumVec2.Current;
      }
    }

    private static double ScalarProduct(IEnumerable<double> vec1, IEnumerable<double> vec2) {
      var enumVec1 = vec1.GetEnumerator();
      var enumVec2 = vec2.GetEnumerator();
      bool f1, f2;
      double product = 0.0;
      while (true) {
        f1 = enumVec1.MoveNext();
        f2 = enumVec2.MoveNext();
        if (!f1 && f2 || f1 && !f2) throw new ArgumentException("vectors are of unequal length.");
        if (!f1 && !f2) return product;
        product += enumVec1.Current * enumVec2.Current;
      }
    }

    private static double VectorLength(IEnumerable<double> vector) {
      return Math.Sqrt(vector.Sum(x => x * x));
    }

    public static NCAClassificationSolution CreateNCASolution(IClassificationProblemData problemData, int k, int reduceDimensions) {
      return new NCAClassificationSolution(problemData, Train(problemData, k, reduceDimensions));
    }

    private class OptimizationInfo {
      public IClassificationProblemData ProblemData { get; set; }
      public int ReduceDimensions { get; set; }
      public OptimizationInfo(IClassificationProblemData problem, int reduceDimensions) {
        this.ProblemData = problem;
        this.ReduceDimensions = reduceDimensions;
      }
    }
  }
}
